import re
from collections import namedtuple
from types import FrameType

from ._internal_debug import debug  # noqa
from .scanner import get_all_blocks
from .scanner import get_variables


class SourceMap:
    
    def __init__(self):
        self._sourcemap = {}
        #   dict[str filepath, dict[int lineno, tuple[str varname, ...]]]
    
    @staticmethod
    def get_frame_info(frame: FrameType, traceback_level: int):
        from .general import normpath
        
        frame_0 = frame
        frame_x: FrameType
        for _ in range(traceback_level):
            frame = frame.f_back
        frame_x = frame
        
        struct = namedtuple('FrameInfo', (
            'direct_filepath', 'direct_lineno',
            'target_filepath', 'target_lineno',
            'target_funcname',
        ))
        
        return struct(
            normpath(frame_0.f_globals.get('__file__', '<unknown>')),
            frame_0.f_lineno,
            normpath(frame_x.f_globals.get('__file__', '<unknown>')),
            frame_x.f_lineno,
            frame_x.f_code.co_name,
        )
    
    def get_sourcemap(self, frame: FrameType, traceback_level: int = 0,
                      advanced=False):
        info = self.get_frame_info(frame, traceback_level)
        
        struct = namedtuple('InfoStruct', (
            'filepath', 'lineno', 'funcname', 'varnames'
        ))
        
        if advanced:
            if info.direct_filepath not in self._sourcemap:
                self._indexing_filemap(info.direct_filepath)
            varnames = self._sourcemap[info.direct_filepath].get(
                info.direct_lineno, ())
        else:
            varnames = ()
        
        return struct(
            info.target_filepath,
            info.target_lineno,
            info.target_funcname,
            varnames,
        )
    
    def _indexing_filemap(self, filepath: str):
        """
        Args:
            filepath
        """
        if filepath.startswith('<'):
            # see `FrameFinder.getinfo._fix_filepath_location`
            self._sourcemap.setdefault(filepath, {})
            return
        
        from .scanner.const import VARIABLE_NAME
        from .scanner.const import SUBSCRIPTABLE
        from .scanner.exceptions import ScanningError
        from .scanner.exceptions import UnresolvedCase
        
        def _get_blocks(lines):
            for match in get_all_blocks(*lines):
                text = match.fulltext.strip()
                if not text:
                    continue
                if not text.startswith('print'):  # FIXME: this is not stable
                    continue
                try:
                    nonlocal node
                    varnames = _analyse_block(text)
                    lineno = match.cursor.lineno + 1
                    node[lineno] = tuple(varnames)
                except ScanningError:
                    nonlocal filepath
                    raise ScanningError(
                        match.cursor.lineno + 1, text,
                        0, '<unknown>',
                        f'<filepath: {filepath}>'
                    )
        
        def _analyse_block(text: str):
            varnames = []
            try:
                for element, type_ in get_variables(text):
                    if type_ == VARIABLE_NAME:
                        varnames.append(element)
                    elif type_ == SUBSCRIPTABLE:
                        varnames.append(_analyse_subscriptables(
                            element, shorten_sub_substrings=True
                        ))
                    else:
                        varnames.append('')
            except UnresolvedCase:
                varnames.clear()
            finally:
                return varnames
        
        def _analyse_subscriptables(
                text: str, shorten_sub_substrings=True, threshold=20
        ):
            """
            shorten_sub_strings:
                example:
                    case 1: when captured a "varname" like "xxx('hello world')":
                        if shorten_sub_strings is True:
                            varname updated: "xxx('...')"
                        if shorten_sub_strings is False:
                            varname updated: "xxx('hello world')"
                    case 2: when captured a "varname" like "xxx('''hello world\n
                            hello world\nhello world\nhello world\nhello...''')"
                            (which is a very long sentense):
                        if shorten_sub_strings is True:
                            varname updated: "xxx('...')"
                        if shorten_sub_strings is False:
                            varname updated: "xxx('''hello world\nhello world\n
                                hello world\nhello world\nhello world\n...)"
                note: the character threshold length to trigger shortening a
                    string is adjustable, the default threshold is 10 chars and
                    must with no line break in it.
            """
            if not shorten_sub_substrings:
                return text
            
            backslash_mask = []
            for i in re.findall(r'\\.', text):
                backslash_mask.append(i)
            if backslash_mask:
                text = re.sub(r'\\.', '__BACKSLASK_MASK__', text)
            
            quotes_pattern_1 = re.compile(r'\'\'\'[\w\W]*\'\'\'|"""[\w\W]*"""')
            quotes_pattern_2 = re.compile(r'\'[^\']*\'|"[^"]*"')
            
            for i in set(quotes_pattern_1.findall(text)):
                if '\n' in i or len(i) > threshold:
                    text = text.replace(i, '"""..."""')
            
            for i in set(quotes_pattern_2.findall(text)):
                if '\n' in i or len(i) > threshold:
                    text = text.replace(i, '"..."')
            
            # restore backslashes
            if backslash_mask:
                for i in backslash_mask:
                    text = text.replace('__BACKSLASK_MASK__', i, 1)
                del backslash_mask
            
            text = re.sub(r'\s+', ' ', text)
            #   note: this ^measure takes a side effect that it may replace
            #   sequential whitespaces to one whitespace inside quote strings.
            
            return text
        
        node = self._sourcemap.setdefault(filepath, {})
        
        with open(filepath, 'r', encoding='utf-8') as f:
            node = self._sourcemap.setdefault(filepath, {})
            _get_blocks((x.rstrip('\n') for x in f.readlines()))


sourcemap = SourceMap()

__all__ = ['sourcemap']



"""
References:
    https://stackoverflow.com/questions/60745841/inheriting-from-baseexception
    -vs-exception
"""


def visualize_line(linex, line, charx, char, symbols=None):
    block = '''
        ┌──────────────────────────────────────────────────────────────────────┐
        │   Tracing linex {}, charx {} (`{}`):
        │       {}
        │       {}
        │   Symbols:
        │       {}
        └──────────────────────────────────────────────────────────────────────┘
    '''.format(
        linex, charx,
        char.replace('\n', '■'),
        line.replace('\n', '■'),
        #   notice: the substitute character width must be ONE. otherwise the
        #   indicator will point to a wrong place.
        ' ' * charx + '^', symbols
    )
    
    temp = []
    for i, line in enumerate(block.splitlines()):
        if i == 0 or i == 7 or i == 8:
            temp.append(line)
        else:
            if len(line) < 80:
                line += ' ' * (80 - len(line) - 1) + '│'
            temp.append(line)
    block = temp

    return '\n'.join(block)


class ScanningError(Exception):
    
    def __init__(self, linex, line, charx, char, symbols):
        self.msg = visualize_line(linex, line, charx, char, symbols)
    
    def __str__(self):
        return self.msg


class UnreachableCase(ScanningError):
    pass


class UnresolvedCase(Exception):
    pass


class UnexpectedReturnCode(Exception):
    pass


class BreakOut(Exception):
    pass



from __future__ import annotations

from textwrap import indent
from typing import Union

from ._internal_debug import debug  # noqa


class MessageFormatter:
    
    @staticmethod
    def markup(*markups: tuple[str, str]) -> str:
        """
        this method produces the final strings that could be directly printed.

        args:
            markups: tuple[tuple[str text, str mark], ...]
                mark:
                    any valid patterns of pytermgui markups. for example, 'red',
                    'bold', 'dim', '#00FF00', ...
                    ps: mark allows empty.
        """
        out = []
        for text, mark in markups:
            if mark:
                out.append('[{}]{}[/]'.format(mark, text))
            else:
                out.append(text)
        return ''.join(out)
    
    # -------------------------------------------------------------------------
    
    def fmt_source(self, filepath: str, lineno: Union[int, str],
                   is_external_lib: bool = False, fmt_width=False) -> str:
        if fmt_width:
            text_a = f'{filepath}:{lineno}'
            text_b = self._fmt_width(text_a, min_width=12)
            additional_space = ' ' * (len(text_b) - len(text_a))
        else:
            additional_space = ''
        if is_external_lib:
            assert filepath.startswith('[')
            a, b = filepath[1:].split(']', 1)
            filepath = self.markup(
                ('\\[{}]'.format(a),
                 'bold {}'.format('red' if filepath.startswith('[unknown]')
                                  else 'magenta')),
                (f'{b}', 'bold blue'),
            )
            return self.markup(
                (filepath, ''),
                (':', 'bright_black'),
                (str(lineno), 'bold blue'),
                (additional_space, ''),
            )
        else:
            return self.markup(
                (filepath.replace('[', '\\['), 'bold blue'),
                (':', 'bright_black'),
                (str(lineno), 'bold blue'),
                (additional_space, ''),
            )
    
    def fmt_separator(self, sep: str = ' >> ', color='bright_black') -> str:
        return self.markup((sep, color))
    
    def fmt_funcname(self, funcname: str, fmt_width=False) -> str:
        is_func = not funcname.startswith('<')
        if is_func:
            funcname += '()'
        else:
            funcname = funcname[1:-1]
        if fmt_width:
            funcname = self._fmt_width(
                funcname, min_width=8, unit_spaces=4
            )
        if is_func:
            return self.markup((funcname, 'green'))
        else:
            return self.markup((funcname, 'magenta'))
    
    def fmt_index(self, idx: Union[int, str]) -> str:
        return self.markup(
            (f'[{idx}]', 'bright_black' if idx == 0 or idx == '0' else 'red')
        )
    
    def fmt_divider(self, div_: str) -> str:
        return self.markup((div_, 'yellow'))
    
    def fmt_message(self, msg_frags: list[str], rich: bool, expand=False,
                    separator=';   ') -> str:  # FIXME
        if not rich:
            msg_frags = (x.replace('[', '\\[') for x in msg_frags)
        if expand:
            return '\n' + indent('\n'.join(msg_frags), '    ')
        else:
            return self.markup((separator, 'bright_black')).join(msg_frags)
    
    def fmt_level(self, text: str, level: str) -> str:
        colors = {
            'trace': '',
            'debug': 'bright_black',
            'info' : 'blue',
            'warn' : 'yellow',
            'error': 'red',
            'fatal': 'bold #ffffff on red',
        }
        return self.markup((text, colors[level]))
    
    # -------------------------------------------------------------------------
    
    @staticmethod
    def _fmt_width(text: str, min_width: int = None, unit_spaces=4) -> str:
        if len(text) <= min_width:
            return text.ljust(min_width)
        else:
            from math import ceil
            return text.ljust(ceil(len(text) / unit_spaces) * unit_spaces)



# TOKENS
TOKEN_A00 = 0x000  # ( <==> )
TOKEN_B00 = 0x001  # ' <==> ' | " <==> " | ''' <==> ''' | """ <==> """
TOKEN_B10 = 0x011  # ' <==> ' | " <==> "
TOKEN_B20 = 0x012  # ''' <==> ''' | """ <==> """
TOKEN_B21 = 0x121  # '''?? | """?? (safe period)
TOKEN_B22 = 0x122  # '''xx | """xx (ridden off safe period)
TOKEN_C00 = 0x003  # \ <==> *
TOKEN_D00 = 0x004  # # <==> ***

# RETURN CODE
INIT = 0
SUBMITTABLE = 1
CONTINUE = 2
BREAK_OUT = 3
UNREACHABLE_CASE = 4

# scanner.py > func:get_variables
VARIABLE_NAME = 0  # varname. e.g. 'a', 'b', 'c'
SUBSCRIPTABLE = 1  # varname with '(' or '[' or '{'. e.g. 'requests.get(...)'
QUOTED_STRING = 2  # qstring. e.g. '"hello world"'
SIMPLE_NUMBER = 3  # simple int or float. e.g. 1, 2, 3, 1.0, 2.0, 3.0, and
#   negative number. but not support 1E10, 1., 2., 3., (they are treated as
#   VARIABLE_NAME)
NESTED_STRUCT = 4  # starts with '(' or '[' or '{'

# PAIRED SYMBOLS
PAIRED_SYMBOLS = {
    '(' : (')', TOKEN_A00),
    '[' : (']', TOKEN_A00),
    '{' : ('}', TOKEN_A00),
    '"' : ('"', TOKEN_B00),
    "'" : ("'", TOKEN_B00),
    '\\': ('*', TOKEN_C00),
    '#' : ('*', TOKEN_D00),
}



from inspect import currentframe

from .general import normpath
from .general import default_print


def debug(*args, condition=True):
    frame = currentframe().f_back
    filepath = normpath(frame.f_globals["__file__"])
    lineno = frame.f_lineno
    source = '{}:{}'.format(filepath, lineno)
    if condition:
        default_print(source, '[LKDEBUG]', *args)



from __future__ import annotations


class LoggingConfig:
    """
    options:
        show_source: bool[true]
            add source info (filepath and line number) prefix to log messages.
            example:
                lk.log('hello world')
                # enabled : './main.py:10  >>  hello world'
                # disabled: 'hello world'
        show_varnames: bool[false]
            show both var names and values. (magic reflection)
            example:
                a, b = 1, 2
                lk.log(a, b, a + b)
                # enabled : 'main.py:11  >>  a = 1; b = 2; a + b = 3'
                # disabled: 'main.py:11  >>  1, 2, 3'
        show_external_lib: bool[true]
            if `param source` came from an external library, whether to print.
            for example, if a third-party library 'xxx' also used `lk.log`,
            its source path (relative to current working dir) may be very long,
            if you don't want to see any prints except your own project, you'd
            set this to False.

        # the following options are only available if `show_external_lib` is
        # true.
        path_style_for_external_lib: literal
            literal:
                'pretty_relpath': default
                    trunscate the source path of external lib to be shorter.
                    example:
                        before:
                            '../../../../site-packages/lk_logger/sourcemap.py'
                            # there may be a lot of '../'.
                        after:
                            '[lk_logger]/sourcemap.py'
                'relpath':
                    a relative path to current working dir. (<- `os.getcwd()`)
                    note there may be a lot of '../../../...' if external lib
                    is far beyond the current working dir.
                'lib_name_only':
                    show only the library name (surrounded by brackets).
                    example: '[lk_logger]'
            ps: if you don't want to show anything, you should turn to set
            `show_external_lib` to False.
    """
    console_width: int | None
    path_style_for_external_lib: str
    rich_traceback: bool
    separator: str
    show_external_lib: bool
    show_source: bool
    show_varnames: bool
    
    _preset_conf = {
        'console_width'              : None,
        'path_style_for_external_lib': 'pretty_relpath',
        'rich_traceback'             : True,
        'separator'                  : ';   ',
        #   suggests: ';   ' | ';\t' | '    ' | ...
        'show_external_lib'          : True,
        'show_source'                : True,
        'show_varnames'              : False,
    }
    
    def __init__(self, **kwargs):
        for k, v in self._preset_conf.items():
            if k in kwargs:
                v = kwargs[k]
            self._apply(k, v)
    
    def update(self, **kwargs):
        for k, v in kwargs.items():
            if k in self._preset_conf and v != getattr(self, k, None):
                self._apply(k, v)
    
    def reset(self):
        for k, v in self._preset_conf.items():
            if v != getattr(self, k, None):
                self._apply(k, v)
    
    def _apply(self, key: str, val: bool | int | str):
        setattr(self, key, val)
        if key == 'console_width':
            if val and isinstance(val, int):
                from .console import console
                console.width = val
        elif key == 'rich_traceback':
            if val:
                # https://rich.readthedocs.io/en/stable/traceback.html
                from rich.traceback import install
                from .console import console
                install(console=console, show_locals=True)
            else:
                pass  # TODO: how to uninstall?



from contextlib import contextmanager

from .const import *
from .symbols import Symbols


class Analyser:
    
    def __init__(self, end='\n'):
        self.end = end
        self.last_char = None
        self.last_last_char = None
        self.safe_period = 0
        self.symbols = Symbols()
    
    def reset(self):
        self.symbols.clear()
    
    @contextmanager
    def trace_index(self, index):
        """
        This is an optional feature. When we use:
            with my_analyser.trace_index(index):
                my_analyser.analyse(some_char)
        The `self.symbols.nested_struct` property will be available to use.
        """
        with self.symbols.trace_index(index):
            yield self
    
    def analyse(self, char):
        last_last_char, last_char = self.last_last_char, self.last_char
        symbols = self.symbols
        ret_code = INIT
        
        if symbols:
            _, target_char, token = symbols.last_symbol
            
            if token == TOKEN_A00:
                if char == target_char:
                    symbols.consume()
                elif char in PAIRED_SYMBOLS:
                    symbols.append(*PAIRED_SYMBOLS[char])
                else:
                    pass
            elif token == TOKEN_B00:
                if char == '\\':
                    symbols.append(*PAIRED_SYMBOLS[char])
                elif last_last_char is not None:
                    if last_char is not None:
                        # cases:
                        #   xxx'
                        #   xxx''
                        #   xxx'xxx
                        if char == target_char:
                            if last_last_char == last_char == char:
                                # case: xxx'' + ' = xxx'''
                                symbols.update(target_char, TOKEN_B21)
                            elif last_char == char:
                                # case: xxx' + ' = xxx''
                                pass
                            else:
                                # case: xxx'xxx + ' = xxx'xxx'
                                symbols.update(target_char, TOKEN_B10)
                                symbols.consume()
                        else:
                            if last_last_char == last_char == target_char:
                                # case: xxx'' + x = xxx''x
                                symbols.update(target_char, TOKEN_B10)
                                symbols.consume(-1)  # MARK: 20210901190803
                                #   notice: here we are consuming LAST_CHAR's
                                #   case, not current char's case. this is the
                                #   ONLY position that current char doesn't get
                                #   consumed, we need to recall `main_analyse`
                                #   to consume it forcely.
                                return self.analyse(char)
                    else:
                        ret_code = UNREACHABLE_CASE
                else:
                    if last_char is not None:
                        # case: '
                        assert last_char == target_char
                        if char == target_char:
                            # case: ' + ' = ''
                            pass
                        else:
                            # case: ' + x = 'x
                            symbols.update(target_char, TOKEN_B10)
                    else:
                        ret_code = UNREACHABLE_CASE
            elif token == TOKEN_B10:
                if char == target_char:
                    symbols.consume()
                elif char == '\\':
                    symbols.append(*PAIRED_SYMBOLS[char])
                else:
                    pass
            elif token == TOKEN_B20:
                # the TOKEN_B2 is a virtual node and has no instances in
                # this method. we could only see its sub node instances
                # (TOKEN_B21, TOKEN_B22).
                ret_code = UNREACHABLE_CASE
            elif token == TOKEN_B21:
                self.safe_period += 1
                if self.safe_period == 1:
                    if char == '\\':
                        self.safe_period = 0
                        symbols.update(target_char, TOKEN_B22)
                        symbols.append(*PAIRED_SYMBOLS[char])
                    else:
                        pass
                elif self.safe_period == 2:
                    self.safe_period = 0
                    symbols.update(target_char, TOKEN_B22)
                    if char == '\\':
                        symbols.append(*PAIRED_SYMBOLS[char])
                    else:
                        pass
                else:
                    ret_code = UNREACHABLE_CASE
            elif token == TOKEN_B22:
                # cases:
                #   xxx'''x'
                #   xxx'''xx
                #   xxx''''x
                #   xxx'''''
                # assert last_last_char is not None
                # assert last_char is not None
                if last_last_char == last_char == char == target_char:
                    # case: xxx''''' + ' = xxx''''''
                    symbols.consume()
                elif char == '\\':
                    symbols.append(*PAIRED_SYMBOLS[char])
                else:
                    pass
            elif token == TOKEN_C00:
                symbols.consume()
            elif token == TOKEN_D00:
                symbols.consume()
                ret_code = BREAK_OUT
            else:
                ret_code = UNREACHABLE_CASE
        else:
            if char in PAIRED_SYMBOLS:
                if char == '#':
                    ret_code = BREAK_OUT
                else:
                    symbols.append(*PAIRED_SYMBOLS[char])
            elif char == self.end:
                ret_code = SUBMITTABLE
        
        self.last_last_char, self.last_char = self.last_char, char
        
        if ret_code == INIT:
            if self.symbols:
                ret_code = CONTINUE
            elif char == self.end:
                ret_code = SUBMITTABLE
            else:
                ret_code = CONTINUE
        return ret_code



from __future__ import annotations

import os
from inspect import currentframe

from ._internal_debug import debug  # noqa
from .console import con_print
from .general import normpath
from .markup import MarkupAnalyser
from .message_formatter import MessageFormatter
from .sourcemap import sourcemap

__all__ = ['LKLogger', 'lk']

formatter = MessageFormatter()
markup_analyser = MarkupAnalyser()


class T:  # Typehint
    from rich.console import RenderableType
    from .markup import TMarks as Marks


class LKLogger:
    
    def __init__(self):
        # self._cache = {}  # TODO: be complete in next version
        # #   dict[tuple k, namedtuple v]
        # #       k: (filepath, lineno)
        # #       v: ...
        from .config import LoggingConfig
        self._config = LoggingConfig()
        self._counter = 0
        self._cwd = normpath(os.getcwd())
        
        # lazy loaded
        self.__external_libs = None
        self.__proj_root = None
    
    def configure(self, clear_old=False, **kwargs) -> None:
        if clear_old:
            self._config.reset()
        self._config.update(**kwargs)
    
    @property
    def _proj_root(self) -> str:
        if self.__proj_root is None:
            from .path_helper import find_project_root
            self.__proj_root = find_project_root()
        return self.__proj_root
    
    @property
    def _external_libs(self) -> dict[str, str]:
        if self.__external_libs is None:
            from .path_helper import indexing_external_libs
            self.__external_libs = indexing_external_libs()
        return self.__external_libs
    
    # -------------------------------------------------------------------------
    
    def log(self, *args, **kwargs) -> None:
        msg = self._build_message(currentframe().f_back, *args)
        # debug(msg)
        con_print(msg, **kwargs)
    
    def fmt(self, *args, **_) -> str:
        return str(self._build_message(currentframe().f_back, *args))
    
    def _build_message(self, frame, *args) -> T.RenderableType:
        global formatter, markup_analyser
        
        info = {
            'arguments'        : [],
            'file_path'        : '',
            'function_name'    : '',
            'index'            : '',
            'is_external_lib'  : False,
            'is_rich_text'     : False,
            'line_number'      : '0',
            'multiple_lines'   : False,
            'log_level'        : '',
            'show_divider_line': False,
            'show_varnames'    : self._config.show_varnames,
            'traceback_level'  : 0,
        }
        
        def analyse_markup() -> tuple[int, T.Marks]:
            nonlocal args
            
            markup_pos: int
            marks: T.Marks
            
            #   0: not exist; 1: the first place; -1: the last place
            is_markup = markup_analyser.is_valid_markup
            if args \
                    and isinstance(args[0], str) \
                    and args[0].startswith(':') \
                    and is_markup(args[0]):
                markup_pos = 1
                markup, args = args[0], args[1:]
            elif len(args) > 1 \
                    and isinstance(args[-1], str) \
                    and args[-1].startswith(':') \
                    and is_markup(args[-1]):
                markup_pos = -1
                markup, args = args[-1], args[:-1]
            else:
                markup_pos = 0
                return markup_pos, {}
            
            assert markup_pos and markup
            
            # analyse markup
            marks = markup_analyser.analyse(markup)
            if not marks:
                raise Exception('Unexpected result from markup analyser',
                                markup, marks)
                # # return markup_pos, {}
            if 'd' in marks:
                info['show_divider_line'] = True
            # TODO: if 'e' in marks: ...  # not implemented yet
            if 'i' in marks:
                # TODO: add total-count feature
                if marks['i'] == 0:  # reset counter
                    self._counter = 0
                    if args:
                        self._counter += 1
                    else:
                        info['arguments'].append('(index reset)')
                else:
                    self._counter += 1
                info['index'] = str(self._counter)  # 0 based
            if 'l' in marks:
                info['multiple_lines'] = True
                from rich.pretty import pretty_repr
                args = tuple(map(pretty_repr, args))
            if 'p' in marks:
                info['traceback_level'] = marks['p']
            if 'r' in marks and 'l' not in marks:
                info['is_rich_text'] = True
            if 's' in marks:
                # s0: disable `show_varnames`
                # s1: disable all, equivalent to `default_print`.
                # note: to make sure rick args take effect, `marks['s']` must
                #   be checked AFTER `marks['r']`.
                if marks['s'] == 0:
                    info['show_varnames'] = False
                else:  # FIXME: this is a temporary solution
                    return markup_pos, {'s': 1}
            if 'v' in marks:
                info['log_level'] = (
                    'trace', 'debug', 'info', 'warn', 'error', 'fatal'
                )[marks['v']]
            
            return markup_pos, marks
        
        markup_pos, marks = analyse_markup()
        if marks.get('s') == 1:
            return (
                '[bright_black]{separator}[/]'.format(
                    separator=self._config.separator
                ).join(map(str, args))
            )
        
        srcmap = sourcemap.get_sourcemap(
            frame=frame,
            traceback_level=info['traceback_level'],
            advanced=info['show_varnames'],
        )
        
        def reformat_arguments_with_varnames(markup_pos: int):
            """
            update the following info:
                info['arguments']
            """
            if args:
                if info['show_varnames']:
                    if markup_pos == 0:
                        varnames = srcmap.varnames
                    elif markup_pos == 1:
                        varnames = srcmap.varnames[1:]
                    else:
                        varnames = srcmap.varnames[:-1]
                    try:
                        assert len(varnames) == len(args), (varnames, args)
                    except AssertionError:
                        # debug('failed extracting varnames')
                        info['arguments'].extend(map(str, args))
                    else:
                        # organize args
                        for n, a in zip(varnames, args):
                            info['arguments'].append(
                                f'{n} = {a}' if n else str(a)
                            )
                else:
                    info['arguments'].extend(map(str, args))
        
        reformat_arguments_with_varnames(markup_pos)
        
        info['function_name'] = srcmap.funcname
        
        def reformat_source_for_external_lib():
            """
            update the following info:
                info['is_external_lib']
                info['file_path']
                info['line_number']
            """
            # show external lib?
            if self._config.show_external_lib:
                # is external lib?
                if self._is_external_lib(srcmap.filepath):  # yes
                    info['is_external_lib'] = True
                    # path format
                    fmt = self._config.path_style_for_external_lib
                    if fmt == 'relpath':
                        info['file_path'] = normpath(
                            os.path.relpath(srcmap.filepath, self._cwd)
                        )
                    else:
                        for lib_path in sorted(
                                self._external_libs, reverse=True
                        ):
                            if srcmap.filepath.startswith(lib_path):
                                lib_name = self._external_libs[lib_path].lower()
                                lib_relpath = (
                                        srcmap.filepath[
                                        len(lib_path):].lstrip('./') or
                                        os.path.basename(srcmap.filepath)
                                )
                                break
                        else:
                            lib_name = ''
                            lib_relpath = srcmap.filepath.lstrip('./')
                        
                        # debug(lib_name, lib_relpath)
                        
                        if fmt == 'pretty_relpath':
                            if lib_name:
                                info['file_path'] = \
                                    '[{}]/{}'.format(lib_name, lib_relpath)
                            else:
                                info['file_path'] = \
                                    '[{}]/{}'.format('unknown', lib_relpath)
                        elif fmt == 'lib_name_only':
                            info['file_path'] = f'[{lib_name}]'
                else:  # no
                    info['file_path'] = normpath(
                        os.path.relpath(srcmap.filepath, self._cwd)
                    )
            else:
                if self._is_external_lib(srcmap.filepath):
                    pass  # info['is_external_lib'] = True ?
                else:
                    info['file_path'] = normpath(
                        os.path.relpath(srcmap.filepath, self._cwd)
                    )
            if (x := info['file_path']) and \
                    not x.startswith(('../', '[')):
                info['file_path'] = './' + x
            
            info['line_number'] = str(srcmap.lineno)
        
        reformat_source_for_external_lib()
        
        # ---------------------------------------------------------------------
        
        # show message
        # noinspection PyListCreation
        message_elements = []
        #   element sequence:
        #       1. source (filename & lineno)
        #       2. funcname
        #       3. log_level
        #       4. index
        #       5. divider_line
        #       6. arguments
        # 1. source
        message_elements.append(
            formatter.fmt_source(
                info['file_path'],
                info['line_number'],
                is_external_lib=info['is_external_lib'],
                fmt_width=True
            )
        )
        message_elements.append(
            formatter.fmt_separator('  >>  ')
        )
        # 2. funcname
        message_elements.append(
            formatter.fmt_funcname(
                info['function_name'],
                fmt_width=True
            )
        )
        message_elements.append(
            formatter.fmt_separator('  >>  ')
        )
        # 3. log_level
        if info['log_level']:
            message_elements.append(
                formatter.fmt_level(
                    text='[{tag}]{sep}'.format(
                        tag=info['log_level'].upper(),
                        sep='\t' if info['log_level'] != 'fatal' else '     '
                        #   to be explained below...
                    ), level=info['log_level']
                )
            )
            r''' about placeholder `sep`:
            
                q: why use '\t'?
                a: '\t' aligns its following message more tidier than spaces.
                   illustration (dot means whitespaces):
                        use ' '                 use '\t'
                        [TRACE].hello world     [TRACE]..hello world
                        [DEBUG].hello world     [DEBUG]..hello world
                        [INFO].hello world      [INFO]...hello world
                        [WARN].hello world      [WARN]...hello world
                        [FATAL].hello world     [FATAL]..hello world
                                                        (↑ looks more pretty)
            
                q: why FATAL doesn't use '\t'?
                a: FATAL's effect is white text on red background.
                   the red background effect will be invalid if we use '\t'.
                   so we have to '     ' instead.
                   the letter number of '   ' depends on the width of its
                   leading content. currently it's `4 * <integer> + 2 + <7 (the
                   length of '[FATAL]'>`. see also `./message_formatter
                   : MessageFormatter : fmt_source : (param) min_width`.
            '''
        # 4. index
        if info['index']:
            message_elements.append(
                formatter.fmt_index(
                    info['index']
                )
            )
            message_elements.append(' ')
        # 5. divider_line
        if info['show_divider_line']:
            message_elements.append(
                # PERF: use auto width?
                formatter.fmt_divider('-' * 80)
            )
            message_elements.append(' ')
        # 6. arguments
        message_elements.append(
            formatter.fmt_message(
                info['arguments'],
                rich=info['is_rich_text'],
                expand=info['multiple_lines'],
                separator=self._config.separator
            )
        )
        if info['log_level'] and \
                info['log_level'] != 'trace':
            message_elements[-1] = formatter.fmt_level(
                message_elements[-1],
                level=info['log_level']
            )
        if info['index'] == '0':
            message_elements[-1] = formatter.markup(
                (message_elements[-1], 'bright_black')
            )
        
        return ''.join(message_elements)
    
    def _is_external_lib(self, filepath: str) -> bool:
        return not filepath.startswith(self._proj_root)


lk = LKLogger()



import traceback
from functools import partial
from functools import wraps

from rich.console import Console as BaseConsole

__all__ = ['con_print', 'con_error', 'console']


class Console(BaseConsole):
    
    def __init__(self):
        super().__init__()
        if self._color_system is None:
            self._color_system = 'standard'
        # TODO (width):
        #   if width longer than default, use single line style; otherwise
        #   split sourcemap and message into different lines.
        pass


console = Console()
con_print = partial(console.print, soft_wrap=True)
con_error = partial(console.print_exception, show_locals=True)


def temporarily_reset_lk_logger(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        import lk_logger
        
        if lk_logger.global_control.STATUS == 'enabled':
            before = lk_logger.unload
            after = lk_logger.enable
        elif lk_logger.global_control.STATUS == 'disabled':
            before = lk_logger.unload
            after = lk_logger.disable
        else:
            before = lambda: None
            after = lambda: None
        
        before()
        out = func(*args, **kwargs)
        after()
        
        return out
    
    return wrapper


setattr(traceback, 'print_exception',
        temporarily_reset_lk_logger(traceback.print_exception))

setattr(traceback, 'print_exc',
        temporarily_reset_lk_logger(traceback.print_exc))



from __future__ import annotations

import os
import sys

from .general import normpath


def find_project_root() -> str:
    """ find personal-like project root.

    proposals:
        1. backtrack from current working dir to find a folder which has
           one of '.idea', '.git', etc. folders.
        2. iterate paths in sys.path, to find a folder which is parent of
           current working dir. (adopted)
           if there are more than one, choose which is shortest.
           if there is none, return current working dir.
    """
    cwd = normpath(os.getcwd())
    paths = tuple(
        x for x in map(normpath, map(os.path.abspath, sys.path))
        if cwd.startswith(x) and os.path.isdir(x)
    )
    if len(paths) == 0:
        return cwd
    elif len(paths) == 1:
        return paths[0]
    else:
        return min(paths, key=lambda x: len(x))


def indexing_external_libs() -> dict[str, str]:
    """
    return:
        dict[str path, str lib_name]
    """
    out = {}
    
    for path in reversed(sys.path):
        if not os.path.exists(path):
            continue
        for root, dirs, files in os.walk(path):
            root = normpath(root)
            out[root] = os.path.basename(root)
            
            for d in dirs:
                if d.startswith(('.', '__')):
                    continue
                if '-' in d or '.' in d:
                    continue
                out[f'{root}/{d}'] = d
            
            # for f in files:
            #     name, ext = os.path.splitext(f)
            #     if ext not in ('.py', '.pyc', '.pyd', '.pyo', '.pyw'):
            #         continue
            #     if '-' in name or '.' in name:
            #         continue
            #     out[f'{root}/{f}'] = name
            break
    
    return out



"""
README:
    `docs/what-is-block.md`
"""
from collections import namedtuple
from re import compile
from string import whitespace

from .analyser import Analyser
from .const import *
from .exceptions import *
from .typehint import *
from .._internal_debug import debug  # noqa


def get_all_blocks(*lines: str, end_mark='\n'):
    """ Get ALL COMPLETE text blocks.
    
    Scanning lines and try to merge the adjacent incomplete lines. If the
    merged result is a complete block (type: `list[str]`), yield it to the
    caller.
    
    Args:
        lines:
        end_mark:
            warnings:
                1. whatever end_mark's value is, it is alwayse be measured as
                   one-character width.
                2. `params:lines:each` mustn't end with end_mark, i.e.
                    assert(all(not x.rstrip().endswith(end_mark)) for x in lines)
    """
    assert all(bool(not x.rstrip().endswith(end_mark)) for x in lines), (
        f'Please make sure each of lines must not end with "{end_mark}". '
        f'(You need to strip them before calling this function)'
    )
    assert len(end_mark) == 1
    
    analyser = Analyser(end=end_mark)
    #   IMPROVEMENT: if `lines:each` contains '\n', which means we can't pass
    #       '\n' to `params:end`, we can use `end=None` instead.
    #       (see reactions at MARK@20210901173115.)
    cursor = Cursor()
    fulltext = end_mark.join(lines)
    last_submit_idx = 0
    
    def _submit():
        nonlocal last_submit_idx
        yield Match(cursor.snapshot(last_submit_idx),
                    fulltext[last_submit_idx:cursor.x],
                    analyser.symbols.matches_nest)
        analyser.reset()
        last_submit_idx = cursor.x + 1
    
    for line in lines:
        cursor.update_lineno()
        
        line_ = line + end_mark
        for char in line_:  # MARK: 20210901173115
            cursor.update_charno()
            # _debug(cursor.i, line, cursor.j, char, analyser.symbols)
            
            with analyser.trace_index(cursor.x):
                ret_code = analyser.analyse(char)
            
            if ret_code == SUBMITTABLE:
                yield from _submit()
                # break
            elif ret_code == CONTINUE:
                continue
            elif ret_code == BREAK_OUT:
                # `BREAK_OUT` is recognized as continuously sending `CONTINUE`
                # command, until we meet an `end_mark` then break.
                while char != end_mark:
                    cursor.update_charno()
                    char = line_[cursor.charno]
                    # debug(cursor.charno, char)
                yield from _submit()
                break
            elif ret_code == UNREACHABLE_CASE:
                raise UnreachableCase(
                    cursor.i, line, cursor.j, char, analyser.symbols
                )
            else:
                raise UnexpectedReturnCode(ret_code)


def get_variables(line: str):
    """

    Workflow:
        1. single_line = 'A, (B, C)'
        2. split by comma: ['A', '(B', 'C)']
        3. analyse each element, try to merge every mergable parts (based on
           pairable symbols): ['A', '(B, C)']
        4. now we know there are two elements in single_line: `A` and `(B, C)`
    """
    caller_pattern = compile(r'^([.\w]+)[(\[{]')  # matching: 'requests.get('
    kwargs_pattern = compile(r'^\w+ *=')  # matching: 'a=1'
    lambda_pattern = compile(r'^lambda ')  # matching: 'lambda *_, **__: ...'
    nested_pattern = compile(r'^[(\[{]')  # matching: '(...', '[...', '{...'
    number_pattern = compile(r'^-?[\d]+(?:\.\d+)?$')  # matching: '123', '-123'
    quotes_pattern = compile(r'^[bfru]*[\'"]')  # matching: '"hello', 'b"hello'
    walrus_pattern = compile(r'^(\w+) *:=')  # mathcing: 'x := 12'
    
    for match0 in get_all_blocks(line):
        start, end = match0.span()  # exterior brackets span
        line = line[start + 1:end].rstrip(whitespace + ',')
        #   `~.rstrip(...)`: for example:
        #       line = 'a, b, c, \n' -> 'a, b, c'
        
        for match1 in get_all_blocks(line, end_mark=','):
            element = match1.fulltext.strip()
            # debug(f'{element = }')
            
            if not element:
                # # continue
                raise ScanningError(
                    match1.cursor.lineno, line,
                    match1.cursor.charno, (line + ',')[match1.cursor.tileno],
                    None
                )
            
            if quotes_pattern.match(element):
                yield element, QUOTED_STRING
            elif number_pattern.match(element):
                yield element, SIMPLE_NUMBER
            elif kwargs_pattern.match(element):
                continue  # continue or break out
            elif nested_pattern.match(element):
                yield element, NESTED_STRUCT
            elif lambda_pattern.match(element):
                ''' TODO (memo)
                
                How it happened?
                    For example:
                        line = 'lambda *args, **kwargs: None'
                        
                    When scanner goes here:
                        lambda *args, **kwargs: None
                                    ^
                    Because 'lambda *args' is a 'complete' part (no unresolved
                    brackets, quotes, etc. left), it treats this comma symbol
                    as an end mark, so scanner thinks it can be submitted and
                    yields 'lambda *args' as an 'element' to the caller.
                    
                    The caller (`lk_logger.sourcemap`) receives 'lambda *args'
                    and '**kwargs: None' one after another, so caller thinks
                    there're two elements found. But when logger tries to
                    demonstrate their number is equivalent (see `lk_logger
                    .logger.format > code:'assert len(info.varnames) ==
                    len(data)'`), an AssertionError is raised.
                
                How to resolve it (in the future)?
                    Before handling this example line, replace the comma with a
                    mask symbol. After handling is over, restore it.
                '''
                raise UnresolvedCase('''
                    We didn't find an ideal way to handle lambda expression
                    without breaking currently designed function.
                    The caller should catch this exception and it has to
                    abandon all its collected varnames which came from this
                    function. Say just take it as nothing received from here.
                ''')
                # see `~/lk_logger/sourcemap.py > class:SourceMap > method:
                # _indexing_filemap`
            elif m := walrus_pattern.match(element):
                yield m.group(1), VARIABLE_NAME
            elif caller_pattern.match(element):
                yield element, SUBSCRIPTABLE
            else:
                yield element, VARIABLE_NAME
        
        break


def _debug(linex, line, charx, char, symbols=None):
    # from lk_logger_3_6 import lk
    # lk.logt('[D4011]', visualize_line(
    #     linex, line, charx, char, symbols), h='parent')
    print(visualize_line(linex, line, charx, char, symbols))


class Match:
    
    def __init__(self, cursor: Union[TCursor, tuple],
                 text: str, matches: TMatches2):
        self.cursor = cursor
        self.fulltext = text
        self._matches = matches
        self.spans = list(matches)
    
    def change_depth(self, depth: int):
        def _recurse(collector, node: dict, current_depth):
            if current_depth < depth:
                for v in node.values():
                    _recurse(collector, v, current_depth + 1)
            elif current_depth == depth:
                collector.extend(node.keys())
            return collector
        
        self.spans = _recurse([], self._matches, 0)
    
    def span(self, idx=0) -> TSpan:
        return self.spans[idx]
    
    def group(self, idx=0):
        start, end = self.span(idx)
        return self.fulltext[start:end + 1]
    
    def groups(self):
        out = []
        for i in range(len(self.spans)):
            out.append(self.group(i))
        return out


class Cursor:
    lineno = -1
    charno = -1
    tileno = -1
    
    _dict: dict  # dict[tileno, tuple[lineno, charno]]
    
    def __init__(self):
        self._dict = {}
        self._snap = namedtuple(
            'CursorShot', ['lineno', 'charno', 'tileno', 'index']
        )
    
    def indexing_fulltext(self, lines):
        for line in lines:
            self.update_lineno()
            for _ in line + '\n':
                self.update_charno()
    
    def update_lineno(self):
        self.lineno += 1
        self.charno = -1
    
    def update_charno(self):
        self.charno += 1
        self.tileno += 1
        self._dict[self.tileno] = (self.lineno, self.charno)
        self._dict[(self.lineno, self.charno)] = self.tileno
    
    @property
    def i(self):
        return self.lineno
    
    @property
    def j(self):
        return self.charno
    
    @property
    def x(self):
        return self.tileno
    
    def trans(self, no: Union[int, tuple]):
        return self._dict[no]
    
    def snapshot(self, index=None):
        if index is None:
            index = self.tileno
        (i, j), x = self.trans(index), index
        return self._snap(
            lineno=i, charno=j, tileno=x, index=index
        )



from __future__ import annotations

import builtins

from ._internal_debug import debug  # noqa
from .general import default_print
from .logger import lk

STATUS = 'unloaded'  # literal['enabled', 'disabled', 'unloaded']
_HAS_WELCOME_MESSAGE_SHOWN = False


def setup(*, quiet=False, clear_pre_configured=False, **kwargs):
    """
    args:
        quiet:
            True: show a welcome message in caller side.
            False: do not show.

            note: the welcome message is shown only once, if caller calls this
                function multi times, only the first time when passes
                `quiet=True` will show this message.
            tip: if you are developing an intermediate/supporting library, it
                is recommended to set `quiet=True`.
        clear_pre_configured:
        kwargs: see `./logger.py > LoggingConfig`.
    """
    global _HAS_WELCOME_MESSAGE_SHOWN
    
    lk.configure(clear_pre_configured, **kwargs)
    setattr(builtins, 'print', lk.log)
    
    if not quiet and not _HAS_WELCOME_MESSAGE_SHOWN:
        _HAS_WELCOME_MESSAGE_SHOWN = True
        
        from random import randint
        color_pairs_group = (
            ('#0a87ee', '#9294f0'),  # calm blue -> light blue
            ('#2d34f1', '#9294f0'),  # ocean blue -> light blue
            ('#ed3b3b', '#d08bf3'),  # rose red -> violet
            ('#f38cfd', '#d08bf3'),  # light magenta -> violet
            ('#f47fa4', '#f49364'),  # cold sandy -> camel tan
        )
        color_pair = color_pairs_group[randint(0, len(color_pairs_group) - 1)]
        slogan = _blend_text('♥ lk-logger is ready', color_pair)
        
        # debug(slogan)
        print(slogan, ':rsp')
    
    global STATUS
    STATUS = 'enabled'


def unload():
    setattr(builtins, 'print', default_print)
    global STATUS
    STATUS = 'unloaded'


def enable():
    setattr(builtins, 'print', lk.log)
    global STATUS
    STATUS = 'enabled'


def disable():
    setattr(builtins, 'print', lambda *_, **__: None)
    global STATUS
    STATUS = 'disabled'


# -----------------------------------------------------------------------------
# neutral functions

def _blend_text(message: str, color_pair: tuple[str, str]):
    """ blend text from one color to another.
    
    source: [lib:rich_cli/__main__.py : blend_text()]
    """
    from rich.color import Color
    from rich.text import Text
    
    color1, color2 = (Color.parse(x).triplet for x in color_pair)
    r1, g1, b1 = color1
    r2, g2, b2 = color2
    dr = r2 - r1
    dg = g2 - g1
    db = b2 - b1
    
    text = Text(message)
    size = len(text)
    
    for index in range(size):
        blend = index / size
        color = '#{}{}{}'.format(
            f'{int(r1 + dr * blend):02X}',
            f'{int(g1 + dg * blend):02X}',
            f'{int(b1 + db * blend):02X}'
        )
        text.stylize(color, index, index + 1)
    return text.markup



from __future__ import annotations

from .general import default_print
from .global_control import disable
from .global_control import enable
from .global_control import setup
from .global_control import unload
from .logger import lk

__version__ = '5.1.5'



from contextlib import contextmanager

from .typehint import *


class Symbols:
    
    def __init__(self):
        self._symbols = []  # type: TSymbols
        self._matches = {}  # type: TMatches1
        self._index = -1
    
    def __str__(self):
        return str(self._symbols)
    
    def __bool__(self):
        return bool(self._symbols)
    
    def __getitem__(self, item):
        return self._symbols[item]
    
    @contextmanager
    def trace_index(self, index):
        self._index = index
        yield self
    
    def consume(self, adjust_offset=0):
        """
        Args:
            adjust_offset: if the caller is not consuming a target character,
                pass it a negative number.
                see `analyser.py > MARK@20210901190803`
        """
        pos_i, _, _ = self._symbols.pop()
        pos_o = self._index + adjust_offset
        self._matches[pos_i] = pos_o
    
    def update(self, char, token):
        index, _, _ = self._symbols[-1]
        self._symbols[-1] = (index, char, token)
    
    def append(self, char, token):
        self._symbols.append((self._index, char, token))
    
    def clear(self):
        self._symbols.clear()
        self._matches.clear()
        self._index = -1
    
    @property
    def last_symbol(self):
        return self._symbols[-1]
    
    @property
    def matches_list(self) -> TSpans:
        return [(k, v) for k, v in self._matches.items()]
    
    @property
    def matches_nest(self) -> TMatches2:
        if not self._matches:
            return {}
        
        consumed = []
        matches = self._matches
        pos_list = sorted(matches.keys())
        
        # noinspection PyUnusedLocal
        def loop(master, node, pos_s, pos_e, start, end):
            for i in range(start, end):
                if i in consumed:
                    continue
                pos_i = pos_list[i]
                pos_o = matches[pos_i]
                # lk.logt('[D1605]', pos_i, pos_o, 'field range', pos_s, pos_e)
                
                if pos_o < pos_e:
                    # lk.logt('[D1427]', 'subset of node', pos_i, pos_o, node)
                    loop(
                        master=node, node=node.setdefault((pos_i, pos_o), {}),
                        pos_s=pos_i, pos_e=pos_o,
                        start=i + 1, end=find_nearest_index(pos_o, i + 1)
                    )
                else:
                    # lk.logt('[D1453]', 'adjacent to node', pos_i, pos_o, master)
                    loop(
                        master=master, node=master.setdefault((pos_i, pos_o), {}),
                        pos_s=pos_i, pos_e=pos_o,
                        start=i + 1, end=find_nearest_index(pos_o, i + 1)
                    )
                consumed.append(i)
        
        def find_nearest_index(pos_e, start):
            for i in range(start, len(pos_list)):
                if pos_list[i] > pos_e:
                    return i
            else:
                return len(pos_list)
        
        loop(
            master=None, node=(out := {}),
            pos_s=0, pos_e=max(matches.values()) + 1,
            start=0, end=len(pos_list)
        )
        return out


symbols = Symbols()



from typing import *

if __name__ == '__main__':
    from lk_logger.scanner.scanner import Match as _Match
    from lk_logger.scanner.scanner import Cursor as _Cursor
else:
    _Match = None
    _Cursor = None

TLine = str
TChar = str

TCursor = _Cursor
TPos = int
TStart = TPos
TEnd = TPos

TToken = int
TSymbols = List[Tuple[TPos, TChar, TToken]]

TMatch = _Match
TSpan = Tuple[TStart, TEnd]
TSpans = List[TSpan]

TMatches1 = Dict[TStart, TEnd]
TMatches2 = Dict[Tuple[TStart, TEnd], Dict]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lk_logger', 'lk_logger.scanner']

package_data = \
{'': ['*']}

install_requires = \
['rich']

setup_kwargs = {
    'name': 'lk-logger',
    'version': '5.1.5',
    'description': 'Python advanced print with varnames.',
    'long_description': '# LK Logger\n\n[中文版](https://blog.csdn.net/Likianta/article/details/124358443)\n\nAdvanced print tool for Python.\n\n![](.assets/20220422181702.png)\n\nFeatures:\n\n-   Show source map, function name and varnames in printing.\n\n    ![](.assets/20220422183248.png)\n\n-   Easy to start. Just add two lines code to enable lk-logger.\n\n    ```python\n    # add this\n    import lk_logger\n    lk_logger.setup()\n\n    # then remain using `print` as usual...\n    print(\'Hello world\')\n    print(123, 456, 789)\n    ...\n    ```\n\n    It will replace Python\'s built-in `print` function to take care all the leftovers.\n\n-   Simple to write:\n\n    ```python\n    # before\n    a, b = 1, 2\n    print(\'a = {}, b = {}, a + b = {}\'.format(a, b, a + b))\n\n    # after\n    print(a, b, a + b)\n    ```\n\n    ![](.assets/20220422182815.png)\n\n    (Another example)\n\n    ![](.assets/20220422183714.png)\n\n-   Non-intrusive. After enable it like above, no more modifications on your source code projects (it is low-cost and low-effort to migrate). You will see the new effects at once.\n\n    This would be a good choice for developers who have dived into their projects with Python `print` to get a new start with a new logging util.\n\n-   Code highlight.\n\n    ![](.assets/20220321155834.png)\n\n## Install\n\n```shell\npip install lk-logger\n```\n\nThe latest version is 5.0.0 or higher.\n\n## Quick Start\n\n```python\nimport lk_logger\nlk_logger.setup(show_varnames=True)\n\nprint(\'hello world\')\n\na, b = 1, 2\nprint(a, b, a + b)\n\nprint(a, b, (c := a + b), c + 3)\n```\n\nScreenshot:\n\n![](.assets/20220321154014.png)\n\n## Advanced Usage\n\nUsually, the above example is enough to use.\n\nThe advanced feature is **"markup"** shorthand.\n\nUse a markup as in the first or the last parameter, the markup is a string that starts with \':\', consists of multiple marks.\n\nFor example:\n\n```python\nprint(\':i\', \'monday\')\nprint(\':i\', \'tuesday\')\nprint(\':i\', \'wednesday\')\n```\n\nIt prints weekdays with a numeric prefix:\n\n![](.assets/20220321155834.png)\n\nAnother one:\n\n```python\nimport lk_logger\nlk_logger.setup()\n\nprint(\'this is a divider\', \':d\')\n\nprint(\':v0\', \'this is a TRACE message\')\nprint(\':v1\', \'this is a DEBUG message\')\nprint(\':v2\', \'this is a INFO  message\')\nprint(\':v3\', \'this is a WARN  message\')\nprint(\':v4\', \'this is a ERROR message\')\nprint(\':v5\', \'this is a FATAL message\')\n```\n\n![](.assets/20220328191717.png)\n\n**Here is a list of all available marks:**\n\n| Mark | Description                                  |\n| :--- | :------------------------------------------- |\n| `:d` | divider line                                 |\n| `:i` | index                                        |\n| `:l` | long / loose format (multiple lines)         |\n| `:p` | parent layer                                 |\n| `:r` | rich format                                  |\n| `:s` | short / single line format                   |\n| `:t` | timestamp (not available in current version) |\n| `:v` | verbosity / log level                        |\n\n**Markup options:**\n\n```\n:d0     default divider line (default)\n:d1+    user defined (if not, fallback to :d0)\n\n:i0     reset index\n:i1     number width fixed to 1 (1, 2, 3, ... 9, 10, 11, ...) (default)\n:i2     number width fixed to 2 (01, 02, 03, ..., 99, 100, 101, ...)\n:i3     number width fixed to 3 (001, 002, 003, ..., 999, 1000, 1001, ...)\n:i4+    number width fixed to *\n:i9+    reserved, not defined yet (will be fallback to :i1)\n\n:l0     let lk-logger decides how to format long message (default)\n:l1     force expand all nodes\n\n:p0     self layer\n:p1     parent layer (default)\n:p2     grand parent layer\n:p3     great grand parent layer\n:p4     great great grand parent layer\n:p5+    great great great ... grand parent layer\n        note: be careful using :p2+, it may crash if the layer not exists\n\n:v0     trace\n        if you don\'t like using number, you can use an alias :vT\n        (:vT is not supported in current version. we\'ll bring it soon)\n:v1     debug (alias is :vD) (default)\n:v2     info (alias is :vI)\n:v3     warning (alias is :vW)\n:v4     error (alias is :vE)\n:v5     fatal (alias is :vF)\n:v6+    user defined (if not, fallback to :v0)\n```\n\n**Detailed examples:**\n\nSee [examples/02_all_markup_usages.py](examples/02_all_markup_usages.py).\n\nScreenshot:\n\n![](.assets/20220422184344.png)\n',
    'author': 'Likianta',
    'author_email': 'likianta@foxmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/likianta/lk-logger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
a general purpose, index based text scanner.

example (conceptual):
    input:
        aaa [bbb (ccc) (ddd] eee
    output:
        - text=aaa , type=plain, span=(0, 4)
        - text=[bbb (ccc) (ddd], type=pair_ab, span=(4, 20)
            - text=bbb , type=plain, span=(4, 8)
            - text=(ccc), type=pair_ab, span=(8, 13)
                - text=ccc, type=plain, span=(9, 12)
            - text= (ddd, type=plain, span=(13, 19)
        - text= eee, type=plain, span=(19, 23)

pip install (optional, for dev only):
    - argsense: a command line argument parser.
    - lk_logger: pretty print with varnames.
"""
from __future__ import annotations

import typing as t
from dataclasses import dataclass
from enum import Enum
from enum import auto as _auto

__all__ = [
    'PRESET_RULES',
    'PairRule',
    'Scanner',
]

# this is a workaround to demolish typehint warning.
auto = lambda: _auto()  # noqa


@dataclass
class Span:
    start_index: int
    start_rowx: int
    start_colx: int
    end_index: int
    end_rowx: int
    end_colx: int


@dataclass
class Segment:
    text: str
    type: str
    span: Span
    pair: tuple[str, str]
    children: list


class PairRule(Enum):
    PLAIN = auto()
    ESCAPE = auto()  # e.g. '\\'
    COMMENT = auto()  # e.g. '#'
    PAIR_AB = auto()  # e.g. '()', '[]', '{}'
    PAIR_AA = auto()  # e.g. '""', "''", '``'
    BLOCK_AA = auto()  # e.g. '```...```'
    BLOCK_AB = auto()  # e.g. '/* ... */'


class State(Enum):
    READY = auto()
    PLAIN = auto()
    PAIRING_START = auto()
    PAIR_START_DONE = auto()
    PAIRING_END = auto()


class T:
    _Start = str
    _End = str
    _Rule = PairRule
    PairRule = t.Iterable[tuple[tuple[_Start, _End], _Rule]]


# PERF: use namedtuple.
class PRESET_RULES:  # noqa
    BASE_RULES = {
        'block_aa.backtick'      : (('```', '```'), PairRule.BLOCK_AB),
        'block_aa.dash'          : (("---", "---"), PairRule.BLOCK_AB),
        'block_aa.double_quotes' : (('"""', '"""'), PairRule.BLOCK_AB),
        'block_aa.single_quotes' : (("'''", "'''"), PairRule.BLOCK_AB),
        'block_ab.slash'         : (('/*', '*/'), PairRule.BLOCK_AB),
        'comment.pound'          : (('#', ''), PairRule.COMMENT),
        'comment.slash'          : (('//', ''), PairRule.BLOCK_AA),
        'escape.backslash'       : (('\\', ''), PairRule.ESCAPE),
        'pair_aa.backtick'       : (('`', '`'), PairRule.PAIR_AA),
        'pair_aa.double_quotes'  : (('"', '"'), PairRule.PAIR_AA),
        'pair_aa.single_quotes'  : (("'", "'"), PairRule.PAIR_AA),
        'pair_ab.braces'         : (('{', '}'), PairRule.PAIR_AB),
        'pair_ab.brackets'       : (('[', ']'), PairRule.PAIR_AB),
        'pair_ab.parentheses'    : (('(', ')'), PairRule.PAIR_AB),
        'pair_ab.square_brackets': (('<', '>'), PairRule.PAIR_AB),
    }
    
    # -------------------------------------------------------------------------
    
    GENERAL_RULES = (
        BASE_RULES['pair_aa.double_quotes'],
        BASE_RULES['pair_aa.single_quotes'],
        BASE_RULES['pair_aa.backtick'],
        BASE_RULES['pair_ab.parentheses'],
        BASE_RULES['pair_ab.brackets'],
        BASE_RULES['pair_ab.braces'],
        BASE_RULES['pair_ab.square_brackets'],
    )
    
    PYTHON_RULES = (
        BASE_RULES['block_aa.double_quotes'],
        BASE_RULES['block_aa.single_quotes'],
        BASE_RULES['comment.pound'],
        BASE_RULES['escape.backslash'],
        BASE_RULES['pair_aa.double_quotes'],
        BASE_RULES['pair_aa.single_quotes'],
        BASE_RULES['pair_ab.parentheses'],
        BASE_RULES['pair_ab.brackets'],
        BASE_RULES['pair_ab.braces'],
    )
    
    JAVASCRIPT_RULES = (
        BASE_RULES['block_aa.backtick'],
        BASE_RULES['block_aa.double_quotes'],
        BASE_RULES['block_aa.single_quotes'],
        BASE_RULES['comment.slash'],
        BASE_RULES['escape.backslash'],
        BASE_RULES['pair_aa.double_quotes'],
        BASE_RULES['pair_aa.single_quotes'],
        BASE_RULES['pair_aa.backtick'],
        BASE_RULES['pair_ab.parentheses'],
        BASE_RULES['pair_ab.brackets'],
        BASE_RULES['pair_ab.braces'],
    )
    
    MARKDOWN_RULES = (
        BASE_RULES['block_aa.backtick'],
        # BASE_RULES['block_aa.dash'],
        BASE_RULES['escape.backslash'],
        BASE_RULES['pair_aa.double_quotes'],
        BASE_RULES['pair_aa.single_quotes'],
        BASE_RULES['pair_aa.backtick'],
        BASE_RULES['pair_ab.parentheses'],
        BASE_RULES['pair_ab.brackets'],
        BASE_RULES['pair_ab.braces'],
        BASE_RULES['pair_ab.square_brackets'],
        (('<!--', '-->'), PairRule.BLOCK_AB),
    )


class Scanner:
    
    def __init__(self, pair_rules: T.PairRule = PRESET_RULES.GENERAL_RULES):
        # print(pair_rules)
        from collections import defaultdict
        self._rules = {x[0][0]: x[1] for x in pair_rules}
        #   dict[str start, rule]
        self._pairs = dict(x[0] for x in pair_rules)
        #   dict[str start, str end]
        # self._pairs_r = {v: k for k, v in self._pairs.items()}
        # assert len(self._pairs) == len(self._pairs_r), (
        #     'text scanner does not support ambiguous pair rules, for example: '
        #     '`[` matches `]` while also `{` matches `]`!'
        # )
        self._pair_initials = defaultdict(lambda: {
            'possible_symbols': [],
            'possible_rules'  : [],
        })
        
        for (start, end), rule in pair_rules:
            self._pair_initials[start[0]]['possible_symbols'].append(start)
            self._pair_initials[start[0]]['possible_rules'].append(rule)
        ''' e.g.
            self._pair_starts = {
                '"': [PairRule.PAIR_AA, PairRule.BLOCK_AA],
                '""': [PairRule.BLOCK_AA],
                '"""': [PairRule.BLOCK_AA],
                ...
            }
        '''
        
        for v in self._pair_initials.values():
            v['possible_symbols'].sort(key=len)
    
    def scan(self, text: str) -> t.Iterator[Segment]:
        pointer = Pointer(text)
        # all indexes are zero-based.
        # last_index = 0
        # curr_index = 0
        goto_index = 0
        over_index = len(text) - 1
        
        def walk(
                start: int, end: int = over_index,
                parent_pair_ends: tuple[str, ...] = (),
                is_block: bool = False,
        ) -> t.Iterator[Segment]:
            
            state = State.READY
            last_index: int
            curr_index: int
            pair_start: str
            pair_end: str
            rule: PairRule = PairRule.PLAIN
            temp_children: list | None = None
            
            def find_line_ending(start: int, end: int = None) -> int:
                nonlocal over_index
                if end is None:
                    end = over_index
                for i in range(start, end + 1):
                    if text[i] == '\n':
                        return i
                else:
                    raise IndexError(
                        'no line ending found in range {} ~ {}:\n'
                        '    {}'.format(start, end, text[start:end])
                    )
            
            def find_new_pair_start(char: str) \
                    -> tuple[str, str, PairRule] | None:
                if x := self._pair_initials.get(char):
                    # 1/2. determine final_symbol
                    final_symbol = ''
                    for symbol in x['possible_symbols']:
                        #   note: x['possible_symbols'] is already sorted by
                        #       symbol length.
                        if match(symbol, char):
                            final_symbol = symbol
                    #   now final_symbol is determined, but it might be empty.
                    #   if empty, it means this is not a valid pair start.
                    
                    # 2/2. get rule
                    if final_symbol:
                        pair_start = final_symbol
                        try:
                            pair_end = self._pairs[pair_start]
                        except KeyError:
                            # usually this exception won't happen. but we play
                            #   a trick that somewhere else may temporarily
                            #   removed `pair_start` from `self._pairs`, thus
                            #   we are encountering an KeyError here.
                            # we can take it as a plain char insteadly.
                            # see also [code](#20220708164828).
                            return None
                        rule = self._rules[final_symbol]
                        return pair_start, pair_end, rule
                else:
                    return None
            
            def match(a: str, b: str) -> bool:
                """
                `b` is equal to `text[curr_index]`. len(b) == 1.
                `a` is a variable length string. len(a) >= 1.
                """
                nonlocal text, curr_index
                if len(a) == 1:
                    return a == b
                else:
                    return a == text[curr_index: curr_index + len(a)]
            
            def reset() -> None:
                nonlocal pair_start, pair_end, rule, state, \
                    temp_children
                state = State.READY
                pair_start = ''
                pair_end = ''
                rule = PairRule.PLAIN
                temp_children = None
            
            def submit(
                    pair: tuple[str, str],
                    rule: PairRule,
                    children: list = None,
                    submit_text: str = None,
                    _curr_index: int = None,
            ) -> Segment:
                nonlocal curr_index, last_index, pointer, text
                if submit_text is None:
                    submit_text = text[last_index: curr_index + 1]
                if _curr_index is None:
                    _curr_index = curr_index
                seg = Segment(
                    text=submit_text,
                    type=rule.name,
                    span=pointer.get_span(last_index, _curr_index),
                    pair=pair,
                    children=children or []
                )
                last_index = _curr_index + 1
                return seg
            
            def submittable(_curr_index: int = None) -> bool:
                nonlocal last_index, curr_index
                if _curr_index is None:
                    _curr_index = curr_index
                return last_index <= _curr_index
            
            # -----------------------------------------------------------------
            
            nonlocal pointer, text
            # nonlocal last_index, curr_index, goto_index, over_index
            nonlocal goto_index, over_index
            
            last_index = start
            curr_index = start
            over_index = end
            
            for curr_index in range(start, end + 1):
                if curr_index < goto_index:
                    # warning: you should make sure there is no '\n' between
                    #   `curr_index` and `goto_index`. i.e. carefully setting
                    #   `goto_index`.
                    continue
                else:
                    goto_index = curr_index
                
                char = text[curr_index]
                # print(curr_index, char, ':v')
                if char == '\n':
                    if is_block:
                        continue
                    else:
                        yield submit(('', ''), PairRule.PLAIN)
                        goto_index = curr_index + 1
                        break
                
                if state == State.READY:
                    if parent_pair_ends and \
                            any(match(x, char) for x in parent_pair_ends):
                        if submittable(curr_index - 1):
                            yield submit(
                                ('', ''), PairRule.PLAIN, [],
                                _curr_index=curr_index - 1
                            )
                        return  # go up to outer recursion.
                    if x := find_new_pair_start(char):
                        pair_start, pair_end, rule = x
                        state = State.PAIRING_START
                        goto_index = curr_index + len(pair_start)
                        continue
                    else:
                        continue
                
                if state == State.PAIRING_START:
                    # for now, we are paring start, but we don't know whether
                    #   it is a valid start. it will clear when paring end is
                    #   done.
                    assert pair_start  # noqa
                    
                    if submittable(idx := curr_index - len(pair_start) - 1):
                        yield submit(
                            ('', ''), PairRule.PLAIN, [],
                            _curr_index=idx
                        )
                    
                    if rule == PairRule.ESCAPE:
                        yield submit(('\\', ''), rule, [], '\\' + char)
                        reset()
                        continue
                    
                    if rule == PairRule.COMMENT:
                        # [#20220708164828] temporarily remove comment symbol
                        #   from self._pairs.
                        temp = self._pairs.pop(pair_start)
                        yield submit(
                            ('#', ''),
                            rule,
                            list(walk(curr_index, )),
                            '#' + text[curr_index: find_line_ending(curr_index)]
                        )
                        # then restore self._pairs.
                        self._pairs[pair_start] = temp
                        reset()
                        break
                    
                    assert pair_end  # noqa
                    if match(pair_end, char):
                        yield submit((pair_start, pair_end), rule, [])
                        reset()
                        continue
                    temp_children = list(walk(
                        curr_index,
                        parent_pair_ends=(pair_end, *parent_pair_ends),
                        is_block=bool(
                            rule in (PairRule.BLOCK_AB, PairRule.BLOCK_AA)
                        )
                    ))
                    state = State.PAIRING_END
                    continue
                
                if state == State.PAIRING_END:
                    assert pair_end and rule  # noqa
                    # rule can never be: ESCAPE, COMMENT
                    if match(pair_end, char):
                        curr_index = curr_index + len(pair_end) - 1
                        goto_index = curr_index + len(pair_end)
                        yield submit(
                            (pair_start, pair_end),
                            rule,
                            temp_children or [],
                        )
                        reset()
                        continue
                    else:
                        rule = PairRule.PLAIN
                        yield submit(
                            ('', ''), rule, [], pair_start,
                            _curr_index=last_index + len(pair_start) - 1
                        )
                        if temp_children:
                            yield from temp_children
                        last_index = curr_index + 1
                        reset()
            
            if rule == PairRule.ESCAPE:
                yield submit(('\\', ''), rule, [], '\\')
            elif rule == PairRule.COMMENT:
                yield submit(('#', ''), rule, [], '#')
            else:
                if submittable():
                    yield submit(('', ''), PairRule.PLAIN, [])
            reset()
        
        def merge_segments(segs: list[Segment]) -> t.Iterator[Segment]:
            seg_0: Segment | None = None
            plain: str = PairRule.PLAIN.name
            for seg_1 in segs:
                if seg_0:
                    if seg_0.type == plain:
                        if seg_1.type == plain:
                            seg_0 = Segment(
                                text=seg_0.text + seg_1.text,
                                type=plain,
                                span=Span(
                                    seg_0.span.start_index,
                                    seg_0.span.start_rowx,
                                    seg_0.span.start_colx,
                                    seg_1.span.end_index,
                                    seg_1.span.end_rowx,
                                    seg_1.span.end_colx
                                ),
                                pair=('', ''),
                                children=[]
                            )
                            continue
                        else:
                            yield seg_0
                    else:
                        yield seg_0
                seg_0 = seg_1
                if seg_1.children:
                    seg_1.children = list(merge_segments(seg_1.children))
            if seg_0:
                yield seg_0
        
        yield from merge_segments(
            list(walk(start=0, end=over_index, is_block=True))
        )


class Pointer:
    _map: dict[int, tuple[int, int]]
    
    #   dict[int index, tuple[int rowx, int colx]]
    
    def __init__(self, text: str):
        self._map = {}
        index = -1
        for rowx, line in enumerate(text.splitlines()):
            for colx, char in enumerate(line):
                index += 1
                self._map[index] = (rowx, colx)
            # the '\n' is counted as an additional `colx`.
            index += 1
            self._map[index] = (rowx, len(line))
    
    def get_span(self, index_0: int, index_1: int) -> Span:
        rowx_0, colx_0 = self._map[index_0]
        rowx_1, colx_1 = self._map[index_1]
        return Span(index_0, rowx_0, colx_0, index_1, rowx_1, colx_1)


if __name__ == '__main__':
    import lk_logger
    from argsense import cli
    
    lk_logger.setup(quiet=True, show_varnames=True)
    scanner = Scanner()
    
    
    def recurse_print_segment(seg: Segment, level=0):
        print('    ' * level +
              '[green]{}[/]'.format(seg.text.replace('\n', '[magenta]\\n[/]')),
              f'[magenta]{seg.pair}[/]',
              f'[red]{seg.type}[/]',
              '[bright_black]{}[/]'.format((seg.span.start_index,
                                            seg.span.end_index)),
              ':rs' + ('i' if level == 0 else ''))
        if seg.children:
            print(':rs', '{spaces}[yellow i]children (#{level}):[/]'.format(
                spaces='    ' * level, level=level + 1))
            for child in seg.children:
                recurse_print_segment(child, level + 1)
    
    
    @cli.cmd()
    def test(text: str, is_file=False):
        """
        args:
            text: a string or a file path.
        kwargs:
            is_file (-f): if True, will treat param `text` as a file path.
        """
        print(':ds', text)
        if is_file:
            path = text
            text = open(path).read()
        for segment in scanner.scan(text):
            recurse_print_segment(segment)
    
    
    cli.run(test)



from typing import TypedDict


class TMarks(TypedDict, total=False):
    d: int
    e: int
    i: int
    l: int
    p: int
    r: int
    s: int
    v: int


class MarkupAnalyser:
    """
    readme:
        ~/docs/markup.zh.md

    markup list:
        :d  divider line
        :e  error
        :i  index
        :l  long / loose format (multiple lines)
        :p  parent layer
        :r  rich
        :s  short / simple / single line
        :t  tag
        :v  verbose / log level
    """
    
    def __init__(self):
        from re import compile
        self._mark_pattern = compile(r'^:([deilprsv][0-9]?)+$')
    
    def is_valid_markup(self, text: str) -> bool:
        return bool(self._mark_pattern.match(text))
    
    @staticmethod
    def analyse(markup: str) -> TMarks:
        """
        return:
            dict[literal mark, int value]
                mark: d, e, i, l, p, r, s, t, v
        """
        out = {}
        defaults = {
            'd': 0,
            'e': 0,
            'i': 1,
            'l': 0,
            'p': 1,
            'r': 0,
            's': 0,
            't': 0,
            'v': 1,
        }
        
        from re import compile
        pattern = compile(r'\w\d?')
        for m in (pattern.findall(markup) or ()):
            if len(m) == 1:
                out[m] = defaults[m]
            else:
                out[m[0]] = int(m[1:])
        return out
